% bpdq_1d_cp : Main implementation of BPDQ algorithm
%
% x_bpdq = bpdq_1d_cp(yq,A,epsilon,p,varargin);
%
% This solves the BPDQ program
% x_bpdq = argmin_x  ||x||_1 s.t. ||yq - A*x||_p <= epsilon
% using the Chambolle-Pock algorithm solving
%       x* = argmin_x f1(Ax) + f2(x)
% with f2 the L1 norm and f1 the indicator 
%
% This is the implementation of the method described in
% "Dequantizing Compressed Sensing : When Oversampling and Non-Gaussian
% Constraints Combine"
% by Laurent Jacques, David Hammond, M. Jalal Fadili
% ( submitted to IEEE Transactions on Signal Processing)
%
% Inputs:
% yq - quantized measurements
% A - sensing matrix
% epsilon - lpball radius, chosen so that original signal is 
%   feasible with high probability
% p - BPDQ moment
%
% Selectable control parameters :
% cp_sigma, cp_tau - algorithm parameters
% cp_maxiter - maximum number of CP iterations CP 
% cp_relerr_thresh - relatize error threshold to terminate CP iteration
% cp_verbosity - 1 to enable output, 0 to supress
%
% Outputs:
% x_bpdq - solution to BPDQ program
% D - debugging structure containing field dr_relerr_save, for
% verifying empirical convergence of algorithm
%
% This file is part of BPDQ Toolbox (Basis Pursuit DeQuantizer)
% Copyright (C) 2009, the BPDQ Team (see the file AUTHORS distributed with
% this library) (See the notice at the end of the file.)

function [x_bpdq,D] = bpdq_1d_cp(yq,A,epsilon,p,varargin)
  control_params={'cp_sigma',1,...
                  'cp_tau',1,...
                  'cp_maxiter',500,...
                  'cp_relerr_thresh',1e-6,...
                  'cp_verbosity',0 ...
                  };
  
  argselectAssign(control_params);
  argselectCheck(control_params,varargin);
  argselectAssign(varargin);
  
  % c is upper frame bound for frame corresponding to matrix A
  % (used for iteration for composition, ref
  % Jacques et. al. Lemma 4)
  % Actually, for Chambolle-Pock algo, we must have sigma*tau <= 1/c.
  
  c = max(svd(A))^2;
  
  if cp_sigma*cp_tau > 1/c,
      error(sprintf('sigma*tau > 1/||A||^2 = %e = 1/%e. Please adapt parameters!', 1/c, c));
  end
  
  [M,N] = size(A);
  opA = opMatrix(A);
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % define prox operators
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
  % f1 is ||yq-x||_p, prox computed iteratively
  % with function composition and lp_ball projection
  % (tgamma may be discarded as f1(x) has only zero and infinite values)
  prox1 = @(u,tgamma) bpdq_proj_lpball_mex(u,yq,epsilon,p);
  
  % f2 is ||x||_1, prox is soft thresholding
  prox2= @(u,tgamma) bpdq_soft_threshold(u,tgamma);
  
  % call Chambolle-Pock iteration with defined prox operators sigma,tau,Th,iter)

  [x_bpdq, relerr_save] = bpdq_cp_iter(opA, prox1, prox2, ...
      cp_sigma, cp_tau, cp_relerr_thresh, cp_maxiter, cp_verbosity);
  
  D.dr_relerr_save=relerr_save;

% The BPDQ Toolbox is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%  
% The BPDQ Toolbox is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%  
% You should have received a copy of the GNU General Public License
% along with The BPDQ Toolbox.  If not, see <http://www.gnu.org/licenses/>.
